moon = {}
spaceship = {}
satellite = {}
asteroids = {}
weapon = {}

function love.load()
	--set the window caption
	love.graphics.setCaption("Asteroids in Space")
	
	--gameMode = 0 - before the start of the game. Ship is invisible
	--gameMode = 1 - playing the game. Ship is visible, can shoot, can be destroyed
	gameMode = 0
	instructions = "Turn A D, Boost W, Fire SPACE"
	musicCredit = "Music 'Twilight' by Stellardrone"
	
	--set countdown time and score
	displayTime = 0
	displayScore = 0
	displayHighScore = "High Score : " .. highScore(0)

	--get the window dimensions
	width = love.graphics.getWidth()
	height = love.graphics.getHeight()
	
	--choose colors to use
	dismenuColor = {120,0,0,255}	--time and score color
	menuColor = {0,120,0,255}		--easy, normal, hard menu color
	normalColor = {255,255,255,255}	--white
	
	--load font for score and timer
	useFontScore = love.graphics.newFont("CONFCRG_.TTF",120)
	useMessages = love.graphics.newFont("arial.ttf",20)
	
	--play some background music
	--STELLARDRONE - http://stellardrone.mymusicstream.com/ - Twilight
	ambientMusic = love.audio.newSource("twilight.mp3","stream")
	ambientMusic:setLooping(true)			--play again when done
	ambientMusic:setVolume(.25)				--lower the volume.  1 is max, 0 is off
	love.audio.play(ambientMusic)			--start playing
	
	--make sounds quieter as they move away from centre of screen
	love.audio.setPosition(width/2,height/2,0)
	love.audio.setDistanceModel("exponent")

	--load an image of the stationary moon and set a position
	moon.image = love.graphics.newImage("moon.gif")
	moon.posX = width-250
	moon.posY = 50
	
	--load a picture of an asteroid. We'll scale for all others
	numAsteroids = 0
	asteroid = love.graphics.newImage("asteroid01.gif")
	
	--load a picture of a bullet
	bullet = love.graphics.newImage("bullet01.gif")
	
	--load asteroid explosion sound
	asteroidExplosion1 = love.audio.newSource("asteroidExplosion.mp3","static")
	asteroidExplosion1:setVolume(.5)
	asteroidExplosion2 = love.audio.newSource("asteroidExplosion.mp3","static")
	asteroidExplosion2:setVolume(.5)
	
	--create a satellite that drifts by
	satellite.image = love.graphics.newImage("satellite.gif")
	satellite.posX = -500
	satellite.posY = 15
	satellite.velX = 25
	satellite.velY = 15
	
	--satellite beeping sound
	satellite.sound = love.audio.newSource("beep.mp3","static")
	satellite.sound:setVolume(1)
	satellite.sound:setLooping(true)
	satellite.sound:setDistance(100,500)
	satellite.sound:play()
	
	--create a spaceship
	spaceship.coasting = love.graphics.newImage("spaceship00.gif")
	spaceship.boost1 = love.graphics.newImage("spaceship01.gif")
	spaceship.boost2 = love.graphics.newImage("spaceship02.gif")
	spaceship.useImage = "coasting"
	spaceship.posX = width/2
	spaceship.posY = height/2
	spaceship.direction = 0
	spaceship.health = 5
	spaceship.acceleration = 250
	spaceship.turnSpeed = 4.5
	spaceship.velX = 0
	spaceship.velY = 0
	
	--spaceship sounds
	spaceship.sound = love.audio.newSource("boost.mp3","static")
	spaceship.shipCollision = love.audio.newSource("shipCollision.mp3","static")
	spaceship.sound:setVolume(.5)
	
	--define weapon
	weapon.sound = love.audio.newSource("shoot.mp3","static")
	weapon.sound:setVolume(.1)
	weapon.sound:setLooping(false)
	weapon.reloadTime = .25
	weapon.speed = 300
	weapon.ready = 0
	weapon.bullets = {}
end

function love.draw()
	--always draw the moon
	love.graphics.draw(moon.image,moon.posX,moon.posY)
	
	if (gameMode == 0) then 
		--in menu at start of game
		--display HighScore, Instructions and Music Credit
		love.graphics.setFont(useMessages)
		love.graphics.print(displayHighScore,width/2-useMessages:getWidth(displayHighScore)/2, 45)
		love.graphics.print(instructions,width/2 - useMessages:getWidth(instructions)/2 ,height-60)
		love.graphics.print(musicCredit,width/2 - useMessages:getWidth(musicCredit)/2,height-30)
	end
	
	--display time left and score
	love.graphics.setFont(useFontScore)
	love.graphics.setColor(dismenuColor)
	love.graphics.print(math.floor(displayTime),10,0)
	love.graphics.print(math.floor(displayScore),10,80)
	
	--healthBar
	love.graphics.rectangle("fill",15,10,math.floor(spaceship.health*205/5),20)
	love.graphics.rectangle("line",15,10,205,20)
	love.graphics.setColor(normalColor)

	if gameMode == 1 then 
		-- playing game so draw spaceship
		love.graphics.draw(spaceship[spaceship.useImage], spaceship.posX, spaceship.posY, spaceship.direction, 1, 1, 21, 36)
	end

	for i=1,#asteroids do
		--draw asteroids regardless of game mode
		love.graphics.draw(asteroid,asteroids[i].posX,asteroids[i].posY,0,asteroids[i].size,asteroids[i].size,72,72)
	end

	for i=1,#weapon.bullets do
		--draw bullets regardless of game mode
		love.graphics.draw(bullet,weapon.bullets[i].posX,weapon.bullets[i].posY,0,1,1,4,4)
	end

	--always draw satellite
	love.graphics.draw(satellite.image, satellite.posX, satellite.posY)

	if gameMode == 0 then 
		--draw menu last so it's on top of everything else
		love.graphics.setFont(useFontScore)
		love.graphics.setColor(menuColor)
		love.graphics.print("EASY",width/2-useFontScore:getWidth("EASY")/2, height/2-120)
		love.graphics.print("NORM",width/2-useFontScore:getWidth("NORM")/2, height/2-40)
		love.graphics.print("HARD",width/2-useFontScore:getWidth("HARD")/2, height/2+40)
		love.graphics.setColor(normalColor)
	end
end

function love.update(dt)
	--update the spaceship and satellite positions
	moveObject(spaceship,dt,45,width,height)
	moveObject(satellite,dt,500,width,height)
	
	--tell sound engine where satellite is so beep can change volume
	satellite.sound:setPosition(satellite.posX, satellite.posY, 0)

	--move bullets.  When they reach the edge of the screen, delete them
	for i=#weapon.bullets,1,-1 do
		flagDelete = moveObject(weapon.bullets[i],dt,0,width,height)
		if flagDelete then
			--moved off edge of screen so delete the bullet
			table.remove(weapon.bullets,i)
		end
	end

	--move asteroids.  Wrap their position when they go more than 500 pixels past the screen
	for i=#asteroids,1,-1 do
		moveObject(asteroids[i],dt,500,width,height)
	end
	
	if gameMode == 1 then 
		--playing the game
		--respond to keys to move the ship and to fire the weapon
		controlSpaceship(dt)
		controlWeapons(dt)
		
		--check whether game is over; spaceship.health = 0 OR displayTime = 0
		if spaceship.health == 0 or math.floor(displayTime) == 0 then
			gameMode = 0
			--compare the current displayScore to the saved highscore and update the display message
			displayHighScore = "High Score : " .. highScore(displayScore) 
		end
		
		--check for collisions between an asteroid and the spaceship
		--walk through the list of asteroids backwards. If go forwards
		--we stand to miss an asteroid if we delete the current.
		for i=#asteroids,1,-1 do
			flagCollide = checkCollision(spaceship,asteroids[i],21+72*asteroids[i].size)
			if flagCollide then 
				if spaceship.shipCollision:isStopped() then 
					--there was a collision and the collision sound isn't playing
					--so reduce the health by 1
					spaceship.health = spaceship.health - 1
					--and play the collision sound
					love.audio.play(spaceship.shipCollision)
				end
			end
		end
		
		--check for collision between bullet and asteroids
		--check each bullet against each asteroid
		for i=#weapon.bullets,1,-1 do
			for j=#asteroids,1,-1 do
				--the collision threshold distance is half a bullet width + half an asteroid width
				flagCollide = checkCollision(weapon.bullets[i],asteroids[j],4+72*asteroids[j].size)
				if flagCollide then 
					--a bullet hit an asteroid
					--destroy the bullet
					--destroy the asteroid if size is < .5
					--split the asteroid into 2 if size is between .5 and .75
					--split the asteroid into 3 if size is between .75 and 1.0
					
					--update the score. Small asteroids give more points
					displayScore = displayScore + math.floor(5/asteroids[j].size)
					
					if asteroids[j].size < 0.5 then
						--nothing to do
					elseif asteroids[j].size < 0.75 then
						--add two new smaller asteroids where current asteroid is
						--speed range is twice as fast as normal
						addAsteroids(2, asteroids[j].posX, asteroids[j].posY,0.5, 2)						
						
					elseif asteroids[j].size <= 1 then
						--add two new smaller asteroids where current asteroid is
						--speed range is twice as fast as normal
						addAsteroids(3, asteroids[j].posX, asteroids[j].posY,0.5, 2)						
					else
						--add four new smaller asteroids where current asteroid is
						--speed range is twice as fast as normal
						addAsteroids(4, asteroids[j].posX, asteroids[j].posY,0.5, 2)						
					end
					
					--remove original asteroid
					table.remove(asteroids,j)
					
					--because asteroid explosions can overlap we'll use two samples
					--if either is free play an explosion
					if asteroidExplosion1:isStopped() then 
						asteroidExplosion1:play()
					elseif asteroidExplosion2:isStopped() then 
						asteroidExplosion2:play()
					end
					--stop checking for asteroid collisions with this bullet
					--since we're going to destroy it in a moment
					break
				end
			end
			
			if flagCollide then
				--destroy the bullet
				table.remove(weapon.bullets,i)
				break
			end
		end
		
	else
		--menu mode
		--wait for easy, norm or hard
		selectGameStart()
	end
	
	if gameMode == 1 then 
		--if the number of asteroids is below the number specified
		--by the difficulty level, make some more.
		if #asteroids < numAsteroids then 
			--if posX and posY are set as nil, the function will randomly generate a position
			addAsteroids(numAsteroids - #asteroids, nil, nil, 1, 1)
		end
		displayTime = displayTime - dt
	end
end

function moveObject(spaceObject,dt, borderWidth, width, height)
	--move an object according to its velX and velY
	--if the object reaches 'borderWidth' past the edge of the screen, wrap it around to the opposite side
	--move object
	
	--function returns true if object was wrapped.  False if not.
	returnValue = false

	--move the object
	spaceObject.posX = spaceObject.posX + spaceObject.velX * dt
	spaceObject.posY = spaceObject.posY + spaceObject.velY * dt
	
	if spaceObject.posX > width + borderWidth then
		--wrap the object from one side of the screen to the other
		returnValue = true
		spaceObject.posX = -borderWidth
		
	elseif spaceObject.posX < -borderWidth then
		--wrap the object from one side of the screen to the other
		spaceObject.posX = width + borderWidth	
		returnValue = true
	end
	
	if spaceObject.posY > height + borderWidth then
		--wrap the object from one side of the screen to the other
		spaceObject.posY = -borderWidth
		returnValue = true
		
	elseif spaceObject.posY < -borderWidth then
		--wrap the object from one side of the screen to the other
		spaceObject.posY = height + borderWidth	
		returnValue = true
	end
	
	return returnValue
end

function controlWeapons(dt)
	--fire the weapon if 'space' is hit and it's reloaded
	if weapon.ready > 0 then
		--when weapon.ready reaches zero, it's ready to fire
		weapon.ready = weapon.ready - dt
	end
	
	if love.keyboard.isDown(" ") then
		if weapon.ready <= 0 then
			--play the firing sound
			if weapon.sound:isStopped() then 
				weapon.sound:play()
			end

			--set the relod time for the next shot
			weapon.ready = weapon.reloadTime
			
			--create a bullet
			table.insert(weapon.bullets,{})
			--position if near the nose of the ship
			--move in the same direction the ship is pointing
			
			--some trigonometry magic
			weapon.bullets[#weapon.bullets].posX = spaceship.posX + math.sin(spaceship.direction) * 35
			weapon.bullets[#weapon.bullets].posY = spaceship.posY - math.cos(spaceship.direction) * 35
			weapon.bullets[#weapon.bullets].velX = math.sin(spaceship.direction) * weapon.speed
			weapon.bullets[#weapon.bullets].velY = -math.cos(spaceship.direction) * weapon.speed
		end
	end
	
end

function controlSpaceship(dt)
	--turn the ship CCW
	if love.keyboard.isDown("a") then 
		spaceship.direction = spaceship.direction - spaceship.turnSpeed * dt
	end
	
	--turn the ship CW
	if love.keyboard.isDown("d") then 
		spaceship.direction = spaceship.direction + spaceship.turnSpeed * dt
	end
	
	--boost the ship
	if love.keyboard.isDown("w") then 
		--turn on the boosting sound if it was off
		if spaceship.sound:isStopped() then 
			spaceship.sound:play()
		end
	
		--some trigonometry magic
		spaceship.velX = spaceship.velX + math.sin(spaceship.direction) * spaceship.acceleration * dt
		spaceship.velY = spaceship.velY + math.cos(spaceship.direction) * -spaceship.acceleration * dt

		--simple animation between two images
		if spaceship.useImage == "boost1" then 
			spaceship.useImage = "boost2"
		else
			spaceship.useImage = "boost1"
		end
	else
		if not spaceship.sound:isStopped() then 
			--turn off the boosting sound if it was on (actually not off)
			spaceship.sound:stop()
		end
		spaceship.useImage = "coasting"

	end
end

function checkCollision(object1,object2,minDistance)
	--use the pythagean thereom to calculate the distance between the two objects
	--distance = SquareRoot(Square(X2 - X1) + Square(Y2 - Y1))
	distance = math.sqrt((object1.posX - object2.posX)^2 + (object1.posY - object2.posY)^2)
	if distance <= minDistance then
		--two objects are closer than the minimum distance so a collision has occurred
		return true
	else
		return false
	end
end

function destroyExistingAsteroids()
	--to keep the code tidy, we'll just destroy the existing asteroids
	for i=#asteroids,1,-1 do
		table.remove(asteroids,i)
	end
end


function addAsteroids(num, posX, posY, maxSize, speedMult)
	--add a number of asteroids
	--if posX or posY are nil then randomly generate off0screen positions
	--otherwise use the supplied position
	--the size of the new asteroids will be between 0.2 and maxSize 
	--the speed of the asteroids will be between +/- maxSpeed
	--speedMult can be used to boost the speed of new asteroids
	maxSpeed = 100
	for i=1,num do
		table.insert(asteroids,{})
		asteroids[#asteroids].size = math.random() * (maxSize - 0.2) + 0.2
		if posX == nil or posY == nil then 
			asteroids[#asteroids].posX = math.random()*(width+1000) - 500
			asteroids[#asteroids].posY = math.random()*(height+1000) - 500
			--random positioned asteroids should appear in the middle of the screen so 
			--shift them off screen
			if asteroids[#asteroids].posX > 0 and asteroids[#asteroids].posX < width then 
				asteroids[#asteroids].posX = asteroids[#asteroids].posX + width
			end
		else
			asteroids[#asteroids].posX = posX
			asteroids[#asteroids].posY = posY
		end
		
		asteroids[#asteroids].velX = (math.random()*2*maxSpeed - maxSpeed) * speedMult
		asteroids[#asteroids].velY = (math.random()*2*maxSpeed - maxSpeed) * speedMult
	end
end

function selectGameStart()
	if love.mouse.isDown("l") then 
		mouseX = love.mouse.getX()
		mouseY = love.mouse.getY()
		if mouseX > 280 and mouseX < 570 and mouseY > 216 and mouseY < 450 then 
			-- pressed a button
			if mouseY > 373 then		-- hard
				numAsteroids = 32
			elseif mouseY > 291 then	-- norm
				numAsteroids = 24
			else						-- easy
				numAsteroids = 16
			end
			--create the initial asteroids
			destroyExistingAsteroids()
			addAsteroids(numAsteroids,nil,nil,1.5,1)
			
			spaceship.posX = width/2
			spaceship.posY = height/2
			spaceship.velX = 0
			spaceship.velY = 0
			spaceship.health = 5
			displayTime = 180
			displayScore = 0
			gameMode = 1
		end
	end
end

function highScore(newScore)

	saveDir = love.filesystem.getSaveDirectory()
	saveFile = "/highscore.txt"
	saveFileExists = love.filesystem.exists(saveFile)  
	if saveFileExists then 
		localHighScore = love.filesystem.read(saveFile)
		localHighScore = tonumber(localHighScore)
	else
		localHighScore = 0
	end
	
	if newScore > localHighScore then 
		localHighScore = newScore
	end
	
	love.filesystem.write(saveFile, localHighScore)
	return localHighScore
end
